__all__ = (
    "Animator",
)

from typing import Any, Sequence, Union

import bpy


class Animator:
    """
    Wrapper around Blender's animation API, providing control of keyframe type and
    handles.

    Usage example:

    .. code-block:: py

       anim = Animator(obj, "location", index=1)   # Y location
       anim.animate(0, 0)
       anim.animate(10, 10, handle="VECTOR")
       anim.animate(20, 20, type="EXTREME")
    """

    obj: bpy.types.Object
    """Blender object to animate."""

    def __init__(self, obj: bpy.types.Object, data_path: str, index: int = 1,
            action_name: str = None):
        """
        :param index: If a vector property, index of the component to animate, e.g. index 1 for Y location.
        :param action_name: Creates new action for obj with this name if not already present.
            Default obj.action
        """
        self.obj = obj
        self.index = index

        if obj.animation_data is None:
            obj.animation_data_create()
        if obj.animation_data.action is None:
            name = action_name if action_name is not None else f"{obj.name}.action"
            obj.animation_data.action = bpy.data.actions.new(name)

        self._fcurve = obj.animation_data.action.fcurves.find(data_path, index=index)
        if self._fcurve is None:
            self._fcurve = obj.animation_data.action.fcurves.new(data_path, index=index)

    @property
    def data_path(self):
        return self._fcurve.data_path

    def animate(self, frame: int, value: Any, handle: str = "AUTO_CLAMPED", type: str = "KEYFRAME"):
        """
        Insert keyframe.
        """
        key = self._fcurve.keyframe_points.insert(frame, value)
        key.handle_right_type = handle
        key.handle_left_type = handle
        key.type = type

        self._fcurve.update()



"""
Procedures relating to intensity of something e.g. light, piano key's rotation,
string's vibration, etc.
"""

__all__ = (
    "Intensity",
    "IntensityOnOff",
    "IntensityFade",
)

import bpy
import numpy as np

from .procedure import Procedure


class Intensity(Procedure):
    """
    Base intensity class.

    Max intensity is determined by note velocity.

    Parameters
    ----------

    animkey: Animation key with following keys:
        - basis: Resting (intensity 0) position.
        - on: Playing (intensity 1) position.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.animkey = kwargs.get("animkey")


class IntensityOnOff(Intensity):
    """
    Turns on when a note starts and off when it ends.

    Keyframe types:
    - JITTER: Resting (intensity = 0).
    - BREAKDOWN: Playing.

    Parameters
    ----------

    duration: Time, in seconds, to spend interpolating from states.
        Default: 0.1

    vector_handles: Whether to use vector handles (no easing in or out).
        Default: False
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.duration = kwargs.get("duration", 0.1)
        self.vector_handles = kwargs.get("vector_handles", False)

    def animate(self):
        duration = self.duration * bpy.context.scene.render.fps
        handle = "VECTOR" if self.vector_handles else "AUTO_CLAMPED"

        for i, note in enumerate(self.midi):
            last = note.prev_end
            next = note.next_start

            keys = []

            # Initial resting position
            if i == 0 or note.start-last > 2*duration:
                keys.append((note.start-duration, 0, "JITTER"))

            # Playing through note
            intensity = np.interp(note.velocity, [0, 127], [0, 1])
            keys.append((note.start, intensity, "BREAKDOWN"))
            keys.append((note.end, intensity, "BREAKDOWN"))

            # Resting after note (sooner if next note close).
            frame = min(note.end+duration, (note.end+next)/2)
            keys.append((frame, 0, "JITTER"))

            for frame, value, type in keys:
                self.animkey.animate(frame, handle=handle, type=type, on=value)


class IntensityFade(Intensity):
    """
    Turns on when a note starts and fades away.

    Keyframe types:
    - JITTER: Resting (intensity = 0).
    - BREAKDOWN: Fading.
    - EXTREME: Max intensity.

    Parameters
    ----------

    fade_fac: Exponential decay factor in factor/sec.
        Default 0.6

    key_interval: Interval in frames for decay section keyframes. Avoids keyframing
        every frame.
        Default: 10

    off_thres: Threshold for intensity to be considered off.
        Default: 0.01

    note_end: Whether to set intensity to 0 at end of note. If false, the note will
        keep glowing until next play.
        Default: True  (stops when note ends).
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fade_fac = kwargs.get("fade_fac", 0.3)
        self.key_interval = kwargs.get("key_interval", 10)
        self.off_thres = kwargs.get("off_thres", 0.01)
        self.note_end = kwargs.get("note_end", True)

    def animate(self):
        fac = self.fade_fac ** (1.0/bpy.context.scene.render.fps)
        interval_fac = fac ** self.key_interval

        # Whether last note ended (intensity reached 0) with enough time to require
        # another keyframe at 0 before next note.
        last_ended = True

        for i, note in enumerate(self.midi):
            last = note.prev_end
            next = note.next_start
            long_pause = next > note.end + 3   # Long between this end and next start
            end_frame = next - 3
            if self.note_end:
                end_frame = min(note.end, end_frame)

            intensity = np.interp(note.velocity, [0, 127], [0, 1])

            # Initial intensity
            if last_ended:
                self.animkey.animate(note.start-1, type="JITTER", on=0)
            self.animkey.animate(note.start, type="EXTREME", on=intensity)

            # Fading
            frame = note.start
            while True:
                intensity *= interval_fac
                next_frame = frame + self.key_interval

                if next_frame > end_frame:
                    final_int = intensity * fac ** (end_frame-frame)
                    self.animkey.animate(end_frame, type="BREAKDOWN", on=final_int)
                    if long_pause:
                        # Need to set to 0
                        self.animkey.animate(end_frame+1, type="JITTER", on=0)
                    break

                if intensity < self.off_thres:
                    self.animkey.animate(next_frame, type="JITTER", on=0)
                    break

                self.animkey.animate(next_frame, type="BREAKDOWN", on=intensity)

                frame = next_frame

            last_ended = long_pause



__all__ = (
    "AnimKey",
)

from typing import Dict, List, Sequence

import numpy as np

from .anim import Animator


class AnimKey:
    """
    Set of value presets for animators.

    Each key is a list of values corresponding to each animator. The basis
    key is always defined at initialize. Other keys are diffs with respect to
    the basis key.

    Animation is done by assigning weights to each key.

    Usage example:

    .. code-block:: python

       anim1 = bmusic.Animator(obj, "location", 0)
       anim2 = bmusic.Animator(obj, "location", 1)
       anim3 = bmusic.Animator(obj, "location", 2)

       key = bmusic.AnimKey([anim1, anim2, anim3], [0, 0, 0])
       key["up"] = (0, 0, 1)
       key["cool"] = (1, 1, -1)
       key.animate(0)
       key.animate(30, up=2, handle="VECTOR", type="JITTER")
       key.animate(60, up=1, cool=2, type="EXTREME")
    """

    _animators: List[Animator]
    _keys: Dict[str, List[float]]

    def __init__(self, animators: List[Animator], basis: Sequence[float]):
        self._animators = animators
        self._keys = {}
        self._keys["basis"] = np.array(basis)

    def __iter__(self):
        return iter(self._keys)

    def __getitem__(self, key: str) -> np.ndarray:
        return self._keys[key]

    def __setitem__(self, key: str, value: Sequence[float]):
        """
        Key is converted to a difference with respect to basis
        i.e. (value - basis)
        """
        assert key not in ("basis", "handle", "type", "frame")
        assert len(value) == len(self._animators)
        value = np.array(value) - self["basis"]
        self._keys[key] = value

    def __delitem__(self, key: str):
        del self._keys[key]

    def __contains__(self, key: str):
        return key in self._keys

    def __len__(self):
        return len(self._keys)

    def animate(self, frame: float, handle: str = "AUTO_CLAMPED", type: str = "KEYFRAME",
            **kwargs):
        """
        Animate all animators.

        :param kwargs: Weight of each shape key.
        """
        value = self["basis"]
        for key, weight in kwargs.items():
            value = value + self[key]*weight

        for i, anim in enumerate(self._animators):
            anim.animate(frame, value[i], handle=handle, type=type)



from typing import Any, Mapping


class Procedure:
    """
    Procedure base class.

    A procedure animates objects, with input parameters from the user.
    Example: Drum mallet animator, vibrating string, etc.

    Extend from this class to create your animator. Subclasses inherit
    available parameters from the parent class.

    Parameters
    ----------

    midi: Midi object.
    """

    params: Mapping[str, Any]

    def __init__(self, **kwargs):
        """
        Initialize the procedure.

        In the subclass, you may override this method, using more parameters.
        Subclasses inherit available parameters from the parent class.
        However, always call super().__init__() to initialize the parent class.
        """
        self.midi = kwargs.get("midi")

    def animate(self):
        """
        Do all animation.
        """
        raise NotImplementedError("Subclass must implement animate()")



import os
import re
import setuptools


with open("bmusic/__init__.py", "r") as fp:
    data = fp.read()

version_pat = re.compile(r"\d\.\d\.\d")
results = version_pat.findall(data)
version = results[0] if len(results) > 0 else "0.0.1"


with open("README.md", "r") as fp:
    long_description = fp.read()

with open("requirements.txt", "r") as fp:
    requirements = fp.read().strip().split("\n")

setuptools.setup(
    name="bmusic",
    version=version,
    author="Patrick Huang",
    author_email="phuang1024@gmail.com",
    description="Utilities for procedural music animation in Blender.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phuang1024/bmusic",
    py_modules=["bmusic"],
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
)



"""
Mechanical movements e.g. hammer, robot arm, etc.
"""

__all__ = (
    "Hammer",
    "Scheduling",
)

from copy import deepcopy
from math import tanh

import bpy
import numpy as np
from tqdm import tqdm

from ..midi import Midi
from .procedure import Procedure


class Hammer(Procedure):
    """
    Hammer movement: Resting, preparing, hitting, recoil, wobbling.

    Keyframe types:
    - JITTER: Resting.
    - BREAKDOWN: Preparing to hit.
    - EXTREME: Hitting.
    - KEYFRAME: Recoil and wobbling.

    Parameters
    ----------

    animkey: Animation key:
        - hit: Hitting.
        - prepare: Preparing to hit.
        - recoil: Bounce back after hit.

    prepare_dur: Duration (sec) of rest to prepare.
        Default: 0.15

    hit_dur: Duration (sec) of prepare to hit movement.
        Default: 0.08

    recoil_dur: Duration (sec) of hit to recoil movement.
        Default: 0.13

    wobble_period: Duration (sec) of each wobble.
        Default: 0.35

    wobble_count: Number of wobbles to perform.
        Default: 4

    wobble_decay: Factor by which wobble intensity decays each time.
        Default: 0.5
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.animkey = kwargs.get("animkey")
        self.prepare_dur = kwargs.get("prepare_dur", 0.15)
        self.hit_dur = kwargs.get("hit_dur", 0.08)
        self.recoil_dur = kwargs.get("recoil_dur", 0.13)
        self.wobble_period = kwargs.get("wobble_period", 0.35)
        self.wobble_count = kwargs.get("wobble_count", 4)
        self.wobble_decay = kwargs.get("wobble_decay", 0.5)

    def animate(self):
        fps = bpy.context.scene.render.fps
        hit_dur = self.hit_dur * fps
        prepare_dur = self.prepare_dur * fps
        recoil_dur = self.recoil_dur * fps
        wobble_period = self.wobble_period * fps
        wobble_count = self.wobble_count
        wobble_decay = self.wobble_decay

        before_dur = prepare_dur + hit_dur
        wobble_dur = wobble_period * wobble_count
        after_dur = wobble_dur + recoil_dur
        total_dur = before_dur + after_dur

        for i, note in enumerate(self.midi):
            last = note.prev_start
            next = note.next_start

            hit_intensity = np.interp(note.velocity, [0, 127], [0, 1])

            # Prepare to hit
            if note.start - last > total_dur:
                # Long time since, so reset to resting position.
                self.animkey.animate(note.start-before_dur, type="JITTER")

            prepare_frame = max(note.start-hit_dur, (note.start+last)/2)
            self.animkey.animate(prepare_frame, type="BREAKDOWN", prepare=hit_intensity)

            # Hit
            self.animkey.animate(note.start, type="EXTREME", handle="VECTOR", hit=hit_intensity)

            # Wobble
            dur_limit = min(after_dur, next-note.start-before_dur)
            offset = recoil_dur

            for j in range(wobble_count):
                if offset >= dur_limit:
                    break
                intensity = hit_intensity * wobble_decay ** (j+1)

                name = "prepare" if j % 2 == 0 else "hit"
                kwargs = {name: intensity}
                self.animkey.animate(note.start+offset, type="KEYFRAME", **kwargs)

                offset += wobble_period

            if next - before_dur > note.start + offset:
                # Long time until, so reset to resting position.
                self.animkey.animate(note.start+offset, type="JITTER")


class Scheduling(Procedure):
    """
    Schedule limited number of objects to switch between more notes to play.
    This procedure only moves the objects, but does not play the notes.
    You may want to combine it with something else e.g. Hammer.

    Parameters
    ----------

    animkeys: List of animation keys, each corresponding to a hammer.
        - move0, move1, move2, ...: Move to note index i.

    idle_time: Time (sec) of pause before moving on to next note.
        Default 0.1

    depth: Depth of search.
        Default: 3
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.animkeys = kwargs.get("animkeys")
        self.idle_time = kwargs.get("idle_time", 0.1)
        self.depth = kwargs.get("depth", 3)

    def animate(self):
        """
        :return: A list. Each element is a Midi obj for the corresponding
            hammer to play.
        """
        self.fps = bpy.context.scene.render.fps  # For performance
        idle_time = self.idle_time * self.fps / 2
        notes_used = self.midi.notes_used

        # (note_ind, last_play_time)
        status = [[None, -1] for _ in self.animkeys]
        schedule = [[] for _ in self.animkeys]

        # Schedule notes
        print("BMusic: Scheduling: Scheduling notes...")
        min_reward = float("inf")
        for i, note in tqdm(enumerate(self.midi), total=len(self.midi.notes)):
            index, reward = self.best_choice(self.midi.notes, i, status, depth=self.depth)

            min_reward = min(min_reward, reward)
            schedule[index].append(note)
            status[index][0] = note.ind
            status[index][1] = note.start

        print(f"BMusic: Scheduling: min_reward={min_reward}")

        # Animate motion
        print("BMusic: Scheduling: Animate motion...")
        midis = list(map(Midi.from_notes, schedule))
        for i, mid in enumerate(midis):
            for note in mid:
                prev = note.prev_start
                next = note.next_start
                frames = [note.start]

                """
                thres = idle_time * 2.5   # Don't want too jarring.
                if note.start-prev > thres:
                    frames.append(note.start-idle_time)
                if next-note.start > thres:
                    frames.append(note.start+idle_time)
                """

                # Midi changed so can't use note.ind
                ind = notes_used.index(note.note)
                kwargs = {f"move{ind}": 1}
                for f in frames:
                    self.animkeys[i].animate(f, **kwargs)

        return midis

    def best_choice(self, notes, note_i, status, depth=1):
        """
        Compute which hammer to choose.

        :return: (hammer_i, reward)
        """
        assert depth >= 1

        count = len(status)
        note = notes[note_i]

        some_not_played = None in [s[0] for s in status]

        reward = []
        for i in range(count):
            if some_not_played:
                # First make sure each hammer plays
                rew = 1e6 if status[i][0] is None else 0
                reward.append(rew)
                continue

            dist = abs(note.ind - status[i][0])
            time = abs(note.start - status[i][1]) / self.fps

            dist = max(dist, 0.1)   # Divide by zero protection
            rew = tanh(30 * time / dist)

            if depth > 1 and note_i < len(notes)-1:
                new_status = deepcopy(status)
                new_status[i][0] = note.ind
                new_status[i][1] = note.start
                _, depth_rew = self.best_choice(notes, note_i+1, new_status, depth-1)

                rew += depth_rew

            reward.append(rew)

        index = np.argmax(reward)
        reward = max(reward)

        return index, reward



from typing import List, Sequence

import bpy
import mido

__all__ = (
    "Note",
    "Midi",
)


class Note:
    """
    Note with absolute start and end time in frames.
    Also supports linked-list like behavior within track.
    """
    note: int
    velocity: int
    start: float
    end: float

    def __init__(self, note, velocity, start, end, midi: "Midi", index: int):
        """
        :param midi: Midi object this note belongs to.
        :param index: Index of this note in the track. Allows for linked-list.
        """
        self.note = note
        self.velocity = velocity
        self.start = start
        self.end = end

        self.midi = midi
        self.index = index

    def __repr__(self):
        return (f"animpiano.Note(note={self.note}, velocity={self.velocity}, "
                f"start={self.start}, end={self.end})")

    def diff(self, other: "Note") -> float:
        """
        Calculate difference in frames between this note and another's start.
        Positive if this note starts after other.
        """
        return self.start - other.start

    @property
    def ind(self) -> int:
        """
        Index of self.note in self.midi.
        """
        return self.midi.notes_used.index(self.note)

    @property
    def next(self) -> "Note":
        """
        Get next note in midi.
        """
        ind = self.index + 1
        if ind >= len(self.midi.notes):
            return None
        return self.midi.notes[ind]

    @property
    def next_start(self) -> float:
        """
        Timestamp of next note's start.
        1e9 if no next note.
        """
        return self.next.start if self.next else 1e9

    @property
    def prev(self) -> "Note":
        """
        Get previous note in midi.
        """
        ind = self.index - 1
        if ind < 0:
            return None
        return self.midi.notes[ind]

    @property
    def prev_start(self) -> float:
        """
        Timestamp of previous note's start.
        -1e9 if no previous note.
        """
        return self.prev.start if self.prev else -1e9

    @property
    def prev_end(self) -> float:
        """
        Timestamp of previous note's end.
        -1e9 if no previous note.
        """
        return self.prev.end if self.prev else -1e9


class Midi:
    """
    Parse a midi file, combining all tracks.
    """

    def __init__(self, path: str = None, offset: float = 0) -> None:
        """
        Parse a midi file.

        :param path: Path to midi file.
        :param fps: Frames per second.
        :param offset: Offset in frames to add to all notes.
        """
        self.notes = []

        if path is None:
            return

        fps = bpy.context.scene.render.fps
        midi = mido.MidiFile(path)

        starts = [0] * 1000
        vels = [0] * 1000

        frame = 0
        started = False
        for msg in midi:
            if started:
                frame += msg.time * fps
    
            if msg.type.startswith("note_"):
                started = True
                note = msg.note
                vel = msg.velocity if msg.type == "note_on" else 0
                if vel == 0:
                    n = Note(note, vels[note], starts[note]+offset, frame+offset, self, len(self.notes))
                    self.notes.append(n)
                else:
                    starts[note] = frame
                    vels[note] = vel

        self._init()

    def __iter__(self):
        return iter(self.notes)

    def __len__(self):
        return len(self.notes)

    @classmethod
    def from_notes(cls, notes: Sequence[Note]) -> "Midi":
        """
        Create a midi from a list of notes.
        """
        midi = Midi(None, None)
        midi.notes = [Note(n.note, n.velocity, n.start, n.end, midi, i) for i, n in enumerate(notes)]
        midi._init()
        return midi

    @property
    def length(self) -> float:
        """
        Duration, in frames, between first note's start and last note's end.
        Does not regard offset.
        """
        return self.notes[-1].end - self.notes[0].start
    
    def filter_notes(self, good: Sequence[int]) -> "Midi":
        """
        Filter notes to only those in good.
        """
        return Midi.from_notes(n for n in self.notes if n.note in good)

    def _init(self):
        """
        More initializing.
        """
        self.notes_used = sorted(set(n.note for n in self.notes))
